from typing import List, Type, Union

from myrabbit import CommandBus, EventBus
from myrabbit.core.consumer.consumer import Consumer, ThreadedConsumer
from myrabbit.core.consumer.listener import Listener
from myrabbit.core.consumer.reconnecting_consumer import ReconnectingConsumer
from myrabbit.core.publisher.reconnecting_publisher import ReconnectingPublisherFactory

from .service import Service
from .service_builder import ServiceBuilder


def _print_motd(services: List[Service]):
    from colorama import init
    from colorama import Fore, Style

    init()

    print(Fore.GREEN + Style.BRIGHT + "Services:" + Style.NORMAL)
    for service in services:
        print("> " + service.service_name)
    print()

    for service in services:
        print(Style.BRIGHT + f"[{service.service_name}]" + Style.NORMAL)

        if (commands := service.doc.get_commands()) :
            print("> Commands:")
            for command in commands:
                print("  - " + command)

        if (events := service.doc.get_events()) :
            print("> Listens to events:")
            for event in events:
                print("  - " + event)

        if (replies := service.doc.get_command_replies()) :
            print("> Handles replies to commands:")
            for reply in replies:
                print("  - " + reply)
        print()

    print(Style.RESET_ALL)


def run_services(
    amqp_url: str,
    *services: Union[Service, ServiceBuilder],
    consumer_cls: Type[Consumer] = ThreadedConsumer,
) -> None:
    factory = ReconnectingPublisherFactory(amqp_url)
    event_bus = EventBus(factory)
    command_bus = CommandBus(factory)

    to_run = []
    for inst in services:
        if isinstance(inst, Service):
            to_run.append(inst)
        elif isinstance(inst, ServiceBuilder):
            to_run.append(inst.build(event_bus, command_bus))
        else:
            raise ValueError(f"Invalid service or builder: {inst!r}")

    _print_motd(to_run)
    listeners: List[Listener] = sum([s.listeners for s in to_run], [])
    consumer = ReconnectingConsumer(
        consumer_cls, consumer_kwargs=dict(amqp_url=amqp_url, listeners=listeners)
    )
    consumer.run()


run_services_threaded = run_services



from typing import Tuple, Type

from pydantic import BaseModel

from .converter import Converter, T


class PydanticConverter(Converter):
    def name_and_body(self, model: T) -> Tuple[str, dict]:
        return model.__class__.__name__, model.dict()  # type: ignore

    def name(self, model: Type[T]) -> str:
        return model.__name__

    def instantiate(self, model_class: Type[T], body: dict) -> T:
        return model_class(**body)  # type: ignore

    def accepts(self, model: T) -> bool:
        try:
            return isinstance(model, BaseModel) or issubclass(model, BaseModel)  # type: ignore
        except TypeError:
            return False



import threading
import typing
from ipaddress import IPv4Address
from typing import Optional
from urllib.parse import urlparse

import pika
from opentelemetry import context, propagators, trace
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from opentelemetry.trace import SpanKind, Status
from opentelemetry.trace.status import StatusCanonicalCode
from pika.channel import Channel
from pika.spec import Basic, BasicProperties

from myrabbit.core import consumer, publisher
from myrabbit.core.consumer.channel import ConsumedChannel


def get_thread_id() -> str:
    return str(threading.current_thread().ident)


def get_thread_name() -> str:
    return threading.current_thread().name


def get_header_from_properties(
    properties: pika.BasicProperties, header_name: str
) -> typing.List[str]:
    headers = properties.headers or {}
    return [value for (key, value) in headers.items() if key == header_name]


class MyrabbitInstrumentor(BaseInstrumentor):
    service_name: str

    def instrument_myrabbit(self, service_name: str, **kwargs):
        self.service_name = service_name
        self.instrument(**kwargs)

    def _instrument(self, **kwargs):
        service_name = self.service_name

        self._original_publish = original_publish = publisher.Publisher.publish
        self._original_handle_message = consumer.Consumer._handle_message
        self._original_handle_message_threaded = (
            consumer.ThreadedConsumer._handle_message
        )

        def instrumented_publish(
            self,
            exchange: str,
            routing_key: str,
            message: bytes,
            properties: Optional[BasicProperties] = None,
        ) -> None:
            original_implementation = original_publish.__get__(self, self.__class__)

            if context.get_value("suppress_instrumentation"):
                original_implementation(exchange, routing_key, message, properties)
                return None

            # Following semantic conventions is used:
            # https://github.com/open-telemetry/opentelemetry-specification/blob/master/specification/trace/semantic_conventions/messaging.md

            exception = None
            span_name = f"{exchange}.{routing_key} send".lower()

            with trace.get_tracer(__name__).start_as_current_span(
                span_name, kind=SpanKind.PRODUCER
            ) as span:
                properties = properties or BasicProperties()

                host = self._connection._impl.params.host
                port = self._connection._impl.params.port
                vhost = self._connection._impl.params.virtual_host

                span.set_attribute("messaging.system", "rabbitmq")
                span.set_attribute("messaging.destination", exchange)
                span.set_attribute("messaging.protocol", "amqp")
                span.set_attribute("messaging.protocol_version", "0.9.1")
                span.set_attribute("messaging.url", f"amqp://{host}:{port}/{vhost}")

                if routing_key:
                    span.set_attribute("messaging.rabbitmq.routing_key", routing_key)

                if properties.correlation_id:
                    span.set_attribute(
                        "messaging.conversation_id", properties.correlation_id
                    )

                try:
                    IPv4Address(host)
                except ValueError:
                    span.set_attribute("net.peer.name", host)
                else:
                    span.set_attribute("net.peer.ip", host)

                span.set_attribute("net.peer.port", port)
                span.set_attribute("net.transport", "IP.TCP")

                span.set_attribute("thread.id", get_thread_id())
                span.set_attribute("thread.name", get_thread_name())

                span.set_attribute("service.name", service_name)

                span.set_attribute("myrabbit.exchange", exchange)
                span.set_attribute("myrabbit.routing_key", routing_key)

                try:
                    span.set_attribute("myrabbit.message", message.decode("utf-8"))
                except UnicodeDecodeError:
                    pass

                headers = properties.headers or {}
                propagators.inject(type(headers).__setitem__, headers)
                properties.headers = headers

                try:
                    original_implementation(exchange, routing_key, message, properties)
                except Exception as exc:
                    exception = exc
                    span.set_status(Status(StatusCanonicalCode.UNKNOWN))
                else:
                    span.set_status(Status(StatusCanonicalCode.OK))

            if exception is not None:
                raise exception.with_traceback(exception.__traceback__)

        def instrumented_handle_message(
            self,
            unused_channel: Channel,
            basic_deliver: Basic.Deliver,
            properties: pika.BasicProperties,
            body: bytes,
            channel: ConsumedChannel,
            original_handle_message,
        ) -> None:
            original_implementation = original_handle_message.__get__(
                self, self.__class__
            )

            if context.get_value("suppress_instrumentation"):
                original_implementation(
                    unused_channel, basic_deliver, properties, body, channel
                )
                return None

            token = context.attach(
                propagators.extract(get_header_from_properties, properties)
            )
            exchange = basic_deliver.exchange
            routing_key = basic_deliver.routing_key
            queue = channel.queue.name
            span_name = f"{exchange}.{routing_key}.{queue} process".lower()

            exception = None

            try:
                with trace.get_tracer(__name__).start_as_current_span(
                    span_name, kind=trace.SpanKind.CONSUMER,
                ) as span:
                    parsed_url = urlparse(self._url)

                    if parsed_url.password:
                        parsed_url = parsed_url._replace(
                            netloc="{}:{}@{}".format(
                                parsed_url.username, "***", parsed_url.hostname
                            )
                        )

                    span.set_attribute("messaging.system", "rabbitmq")
                    span.set_attribute("messaging.destination", exchange)
                    span.set_attribute("messaging.operation", "process")
                    span.set_attribute("messaging.protocol", "amqp")
                    span.set_attribute("messaging.protocol_version", "0.9.1")
                    span.set_attribute("messaging.url", parsed_url.geturl())
                    span.set_attribute("messaging.message_id", properties.message_id)

                    if routing_key:
                        span.set_attribute(
                            "messaging.rabbitmq.routing_key", routing_key
                        )

                    if properties.correlation_id:
                        span.set_attribute(
                            "messaging.conversation_id", properties.correlation_id
                        )

                    try:
                        IPv4Address(parsed_url.hostname)
                    except ValueError:
                        span.set_attribute("net.peer.name", parsed_url.hostname)
                    else:
                        span.set_attribute("net.peer.ip", parsed_url.hostname)

                    span.set_attribute("net.peer.port", parsed_url.port)
                    span.set_attribute("net.transport", "IP.TCP")

                    span.set_attribute("thread.id", get_thread_id())
                    span.set_attribute("thread.name", get_thread_name())

                    span.set_attribute("service.name", service_name)

                    span.set_attribute("myrabbit.exchange", exchange)
                    span.set_attribute("myrabbit.routing_key", routing_key)

                    try:
                        span.set_attribute("myrabbit.message", body.decode("utf-8"))
                    except UnicodeDecodeError:
                        pass

                    try:
                        original_implementation(
                            unused_channel, basic_deliver, properties, body, channel
                        )
                    except Exception as exc:
                        exception = exc
                        span.set_status(Status(StatusCanonicalCode.UNKNOWN))
                    else:
                        span.set_status(Status(StatusCanonicalCode.OK))
            finally:
                context.detach(token)

            if exception is not None:
                raise exception.with_traceback(exception.__traceback__)

        def _wrap(impl):
            def inner(*args, **kwargs):
                return instrumented_handle_message(
                    *args, **kwargs, original_handle_message=impl
                )

            return inner

        publisher.Publisher.publish = instrumented_publish
        consumer.Consumer._handle_message = _wrap(consumer.Consumer._handle_message)
        consumer.ThreadedConsumer._handle_message = _wrap(
            consumer.ThreadedConsumer._handle_message
        )

    def _uninstrument(self, **kwargs):
        publisher.Publisher.publish = self._original_publish
        consumer.Consumer._handle_message = self._original_handle_message
        consumer.ThreadedConsumer._handle_message = (
            self._original_handle_message_threaded
        )



import logging
import uuid
from functools import wraps
from typing import Any, Callable, Optional, Union

from pika import BasicProperties

from myrabbit.commands import command_outcome
from myrabbit.commands.command_outcome import CommandReply
from myrabbit.commands.command_with_message import CommandWithMessage, ReplyWithMessage
from myrabbit.commands.reply_headers import CommandReplyHeaders
from myrabbit.core.consumer.callbacks import Callbacks
from myrabbit.core.consumer.listener import Exchange, Listener, Queue
from myrabbit.core.consumer.pika_message import PikaMessage
from myrabbit.core.consumer.reply import Reply
from myrabbit.core.publisher.reconnecting_publisher import PublisherFactory
from myrabbit.core.serializer import JsonSerializer, Serializer

logger = logging.getLogger(__name__)


class CommandBus:
    def __init__(
        self,
        publisher_factory: PublisherFactory,
        serializer: Optional[Serializer] = None,
        default_exchange_params: Optional[dict] = None,
        default_queue_params: Optional[dict] = None,
        callbacks: Optional[Callbacks] = None,
    ):
        self._publisher_factory = publisher_factory
        self._serializer: Serializer = serializer or JsonSerializer()
        self.default_exchange_params = default_exchange_params or {}
        self.default_queue_params = default_queue_params or {}
        self._callbacks = callbacks

    def set_callbacks(self, callbacks: Callbacks) -> None:
        self._callbacks = callbacks

    def send(
        self,
        command_sender: str,
        command_destination: str,
        command_name: str,
        body: Optional[dict] = None,
        properties: Optional[BasicProperties] = None,
        reply_headers: Optional[dict] = None,
    ) -> None:
        properties = properties or BasicProperties()

        body = body or {}
        content_type, binary_body = self._serializer.serialize(body)

        properties.content_type = content_type

        if properties.reply_to is None:
            properties.reply_to = self.reply_queue_name(
                command_sender, command_destination, command_name
            )

        if properties.correlation_id is None:
            properties.correlation_id = uuid.uuid4().hex

        if reply_headers:
            self._set_reply_headers(properties, reply_headers)

        with self._publisher_factory.publisher() as publisher:
            publisher.publish(
                self._exchange(command_destination),
                self._routing_key(command_name),
                binary_body,
                properties,
            )

    def _set_reply_headers(
        self, properties: BasicProperties, reply_headers: dict
    ) -> None:
        properties.headers = properties.headers or {}
        properties.headers.update(reply_headers)
        properties.headers[CommandReplyHeaders.REPLY_HEADERS_KEY] = ",".join(
            reply_headers
        )

    def _get_reply_headers(self, incoming_headers: Optional[dict]) -> dict:
        incoming_headers = incoming_headers or {}
        reply_headers = {}
        reply_to_headers = incoming_headers.get(
            CommandReplyHeaders.REPLY_HEADERS_KEY, ""
        ).split(",")
        for header_name in reply_to_headers:
            if header_name in incoming_headers:
                reply_headers[header_name] = incoming_headers[header_name]
        return reply_headers

    def listener(
        self,
        command_destination: str,
        command_name: str,
        callback: Callable[[CommandWithMessage], Optional[Union[CommandReply, Any]]],
        exchange_params: Optional[dict] = None,
        queue_params: Optional[dict] = None,
    ) -> Listener:
        queue_params = queue_params or {}
        queue_params = {**self.default_queue_params, **queue_params}
        queue_params.setdefault(
            "name", f"{command_destination}.{command_name}",
        )

        exchange_params = exchange_params or {}
        exchange_params = {**self.default_exchange_params, **exchange_params}
        exchange_params.setdefault("type", "direct")
        exchange_params.setdefault("name", self._exchange(command_destination))

        @wraps(callback)
        def deserialize_command_and_handle_reply(
            message: PikaMessage,
        ) -> Optional[Reply]:
            reply_headers = self._get_reply_headers(message.properties.headers)

            try:
                callback_result: Optional[Union[CommandReply, Any]] = callback(
                    CommandWithMessage(
                        self._serializer.deserialize(message.body), message
                    )
                )
            except Exception as e:
                logger.exception(
                    "Exception happened during handling message %s using handler %s",
                    message,
                    callback,
                )
                return (
                    command_outcome.exception(e)
                    .with_headers(reply_headers)
                    .to_reply(self._serializer)
                )

            if callback_result is None:
                return None

            if isinstance(callback_result, CommandReply):
                return callback_result.with_headers(reply_headers).to_reply(
                    self._serializer
                )

            return (
                command_outcome.success(body=callback_result)
                .with_headers(reply_headers)
                .to_reply(self._serializer)
            )

        return Listener(
            exchange=Exchange(**exchange_params),
            queue=Queue(**queue_params),
            routing_key=self._routing_key(command_name),
            handle_message=deserialize_command_and_handle_reply,
            callbacks=self._callbacks,
        )

    def reply_listener(
        self,
        command_sender: str,
        command_destination: str,
        command_name: str,
        callback: Callable[[ReplyWithMessage], None],
        exchange_params: Optional[dict] = None,
        queue_params: Optional[dict] = None,
    ) -> Listener:
        queue_params = queue_params or {}
        queue_params = {**self.default_queue_params, **queue_params}
        queue_params.setdefault(
            "name",
            self.reply_queue_name(command_sender, command_destination, command_name),
        )
        routing_key = queue_params["name"]  # Exchange is direct.

        exchange_params = exchange_params or {}
        exchange_params = {**self.default_exchange_params, **exchange_params}
        exchange_params.setdefault("type", "direct")
        exchange_params.setdefault("name", self._exchange(command_destination))

        @wraps(callback)
        def deserialize_message(message: PikaMessage) -> None:
            reply: dict = self._serializer.deserialize(message.body)
            callback(ReplyWithMessage(reply=reply, message=message))

        return Listener(
            exchange=Exchange(**exchange_params),
            queue=Queue(**queue_params),
            routing_key=routing_key,
            handle_message=deserialize_message,
            callbacks=self._callbacks,
        )

    def _exchange(self, command_destination: str) -> str:
        return f"{command_destination}.commands"

    def _routing_key(self, command_name: str) -> str:
        return command_name

    def reply_queue_name(
        self, command_sender: str, command_destination: str, command_name: str
    ) -> str:
        return f"{command_sender}.listen-reply-to:{command_destination}.{command_name}"



from typing import Callable, Optional, Type

from myrabbit import CommandBus, EventBus
from myrabbit.commands.command_with_message import CommandReplyType, CommandType, ReplyWithMessage
from myrabbit.core.consumer.callbacks import Callback, Middleware
from myrabbit.events.event_with_message import EventType
from myrabbit.events.listen_event_strategy import ListenEventStrategy

from .service import Service


class ServiceBuilder:
    def __init__(self, service_name: str):
        self.service_name = service_name
        self._calls = []

    def before_request(self, fn: Callback) -> Callback:
        self._calls.append(lambda s: s.before_request(fn))
        return fn

    def after_request(self, fn: Callback) -> Callback:
        self._calls.append(lambda s: s.after_request(fn))
        return fn

    def middleware(self, fn: Middleware) -> Middleware:
        self._calls.append(lambda s: s.middleware(fn))
        return fn

    def on_event(
        self,
        event_source: str,
        event_type: Type[EventType],
        exchange_params: Optional[dict] = None,
        queue_params: Optional[dict] = None,
        listen_strategy: Optional[ListenEventStrategy] = None,
        method_name: Optional[str] = None,
    ) -> Callable:
        def register_event_listener(fn):
            self._calls.append(
                lambda s: s.on_event(
                    event_source,
                    event_type,
                    exchange_params,
                    queue_params,
                    listen_strategy,
                    method_name,
                )(fn)
            )
            return fn

        return register_event_listener

    def on_command(
        self,
        command_type: Type[CommandType],
        exchange_params: Optional[dict] = None,
        queue_params: Optional[dict] = None,
    ) -> Callable:
        def register_command_listener(fn: Callable) -> Callable:
            self._calls.append(
                lambda s: s.on_command(command_type, exchange_params, queue_params)(fn)
            )
            return fn

        return register_command_listener

    def on_command_reply(
        self,
        command_destination: str,
        command_type: Type[CommandType],
        reply_type: Optional[Type[CommandReplyType]] = None,
        listen_on: Optional[str] = None,
        exchange_params: Optional[dict] = None,
        queue_params: Optional[dict] = None,
    ) -> Callable:
        def register_command_listener(
            fn: Callable[[ReplyWithMessage], None]
        ) -> Callable:
            self._calls.append(
                lambda s: s.on_command_reply(
                    command_destination,
                    command_type,
                    reply_type,
                    listen_on,
                    exchange_params,
                    queue_params,
                )(fn)
            )
            return fn

        return register_command_listener

    def build(self, event_bus: EventBus, command_bus: CommandBus):
        s = Service(self.service_name, event_bus, command_bus)
        for call in self._calls:
            call(s)
        return s



class CommandReplyHeaders:
    COMMAND_REPLY_PREFIX = "X-Command-Reply"
    REPLY_OUTCOME: str = f"{COMMAND_REPLY_PREFIX}-Outcome"
    REPLY_HEADERS_KEY: str = f"{COMMAND_REPLY_PREFIX}-Headers"
    REPLY_TYPE: str = f"{COMMAND_REPLY_PREFIX}-Type"



import abc

import pika

from myrabbit.core.publisher import Publisher


class PublisherFactory(abc.ABC):
    @abc.abstractmethod
    def get_connection(self) -> pika.BlockingConnection:
        pass

    @abc.abstractmethod
    def publisher(self) -> Publisher:
        pass


class ReconnectingPublisherFactory(PublisherFactory):
    def __init__(self, amqp_url: str):
        self._amqp_url = amqp_url

    def get_connection(self) -> pika.BlockingConnection:
        # TODO: implement reconnection on expected exceptions
        #   and reconnection strategy (block, retry times, delays).
        parameters = pika.URLParameters(self._amqp_url)
        return pika.BlockingConnection(parameters)

    def publisher(self) -> Publisher:
        return Publisher(self.get_connection())



import inspect
from functools import partial
from typing import Callable, Optional


class MethodNameError(Exception):
    pass


def get_method_name(method_name: Optional[str], fn: Callable) -> str:
    if isinstance(fn, partial):
        fn = fn.func

    if inspect.ismethod(fn.__call__):
        return method_name or fn.__class__.__name__

    if method_name:
        return method_name

    try:
        return fn.__name__
    except AttributeError:
        raise MethodNameError(
            f"Can not get method name for object {fn!r}. "
            f"Please, specify method name explicitly.  "
        )



from functools import wraps
from typing import Any, Callable, List, Optional, Type, Union

from pika import BasicProperties

from myrabbit.commands.command_bus import CommandBus
from myrabbit.commands.command_outcome import CommandReply
from myrabbit.commands.command_with_message import (
    CommandReplyType,
    CommandType,
    CommandWithMessage,
    ReplyWithMessage,
)
from myrabbit.core.consumer.listener import Listener
from myrabbit.core.converter import DEFAULT_CONVERTERS, Converter


class CommandBusAdapter:
    def __init__(
        self, command_bus: CommandBus, converters: Optional[List[Converter]] = None,
    ):
        self.command_bus = command_bus
        self.converters = converters or DEFAULT_CONVERTERS

    def get_converter(self, command: CommandType) -> Converter:
        for converter in self.converters:
            if converter.accepts(command):
                return converter
        raise RuntimeError("Converter not found")

    def send(
        self,
        command_sender: str,
        command_destination: str,
        command: CommandType,
        properties: Optional[BasicProperties] = None,
        reply_headers: Optional[dict] = None,
    ) -> None:
        command_name, body = self.get_converter(command).name_and_body(command)
        self.command_bus.send(
            command_sender,
            command_destination,
            command_name,
            body,
            properties,
            reply_headers,
        )

    def listener(
        self,
        command_destination: str,
        command_type: Type[CommandType],
        callback: Callable[[CommandWithMessage], Optional[Union[CommandReply, Any]]],
        exchange_params: Optional[dict] = None,
        queue_params: Optional[dict] = None,
    ) -> Listener:
        converter = self.get_converter(command_type)

        @wraps(callback)
        def instantiate_command(
            command_with_message: CommandWithMessage,
        ) -> Optional[Union[CommandReply, Any]]:
            command_with_message.command = converter.instantiate(
                command_type, command_with_message.command
            )
            return callback(command_with_message)

        command_name = converter.name(command_type)

        return self.command_bus.listener(
            command_destination=command_destination,
            command_name=command_name,
            callback=instantiate_command,
            exchange_params=exchange_params,
            queue_params=queue_params,
        )

    def reply_listener(
        self,
        command_sender: str,
        command_destination: str,
        command_type: Type[CommandType],
        callback: Callable[[ReplyWithMessage], None],
        reply_type: Optional[Type[CommandReplyType]] = None,
        exchange_params: Optional[dict] = None,
        queue_params: Optional[dict] = None,
    ) -> Listener:
        @wraps(callback)
        def instantiate_reply(reply: ReplyWithMessage) -> None:
            if reply_type is not None:
                converter = self.get_converter(reply_type)
                reply.reply = converter.instantiate(reply_type, reply.reply)
            callback(reply)

        command_name = self.get_converter(command_type).name(command_type)

        return self.command_bus.reply_listener(
            command_sender=command_sender,
            command_destination=command_destination,
            command_name=command_name,
            callback=instantiate_reply,
            exchange_params=exchange_params,
            queue_params=queue_params,
        )



from .commands import CommandBus, CommandBusAdapter, CommandWithMessage
from .core.consumer.listener import Listener
from .core.consumer.pika_message import PikaMessage
from .core.publisher.reconnecting_publisher import PublisherFactory, ReconnectingPublisherFactory
from .events import EventBus, EventBusAdapter, EventWithMessage
from .service import Service, ServiceBuilder, run_services, run_services_threaded
